/*
the train and test split could also be made while labelling - i find it more convenient
to build a data set first, then save it and finally split and train
so the example assumes you have your in- and out-datasets saved to disk already
*/

/*
i have some data to use:
7465 points (from audio slices)
inputVec 17 dimens: 13MFCCs, flux(mean), specShapes: centroid,flatness,crest (all mid)
outputVec 1 dimension
all normalized
identifier names are "0", "1" ...
*/

(
~in = FluidDataSet(s).read("".resolveRelative+/+"/datasets/inputVec.json");
~inCopy = FluidDataSet(s).read("".resolveRelative+/+"/datasets/inputVec.json"); //copy needed later
~out = FluidDataSet(s).read("".resolveRelative+/+"/datasets/outputVec.json");
~outCopy = FluidDataSet(s).read("".resolveRelative+/+"/datasets/outputVec.json");
~in.size(action:{arg size; ~size = size}); //gives the size (same for both)
//~size;
)


// rand split
(
~randIds = Array.series(~size).scramble; //build random ids
~split = 0.025; //split for test or train (percentage as decimal)

~testIds = ~randIds[0..(~size*~split).floor.asInteger]; // 2.5% test set
~trainIds = ~randIds[(~size*~split).floor.asInteger..~size][1..]; // 97.5% train set (but first (else in test and train))

~testIds.do({arg id; ~in.deletePoint(id.asString)});
~in.write("".resolveRelative+/+"/datasets/inputVecTrain.json");
~trainIds.do({arg id; ~inCopy.deletePoint(id.asString)});
~inCopy.write("".resolveRelative+/+"/datasets/inputVecTest.json");

~testIds.do({arg id; ~out.deletePoint(id.asString)});
~out.write("".resolveRelative+/+"/datasets/outputVecTrain.json");
~trainIds.do({arg id; ~outCopy.deletePoint(id.asString)});
~outCopy.write("".resolveRelative+/+"/datasets/outputVecTest.json");
)


(
// bring it in
~inTrain = FluidDataSet(s).read("".resolveRelative+/+"/datasets/inputVecTrain.json");
~inTest = FluidDataSet(s).read("".resolveRelative+/+"/datasets/inputVecTest.json");

~outTrain = FluidDataSet(s).read("".resolveRelative+/+"/datasets/outputVecTrain.json");
~outTest = FluidDataSet(s).read("".resolveRelative+/+"/datasets/outputVecTest.json");

~inNormTrain = FluidNormalize(s).fitTransform(~inTrain,~inTrain);
~inNormTest = FluidNormalize(s).fitTransform(~inTest,~inTest);
)


~inTest.size(action:{arg size; ~tsize = size}); //get the size of the testsplit (we have 187 to use to test)
~tsize; //to choose a working test batch size

(
// a model
~mlp = FluidMLPRegressor(s,[8,5,3],FluidMLPRegressor.sigmoid,FluidMLPRegressor.sigmoid,maxIter:100,learnRate:0.1,batchSize:16,validation:0.1);

~continuously_train = true;

// empty lists to populate with train- and testloss vals (validation)
~trainLoss = List.new;
~testLoss = List.new;

~predSourceBuf = Buffer.loadCollection(s,[0]);
~predTargetBuf = Buffer.loadCollection(s,[0]);
~groundTruthBuf = Buffer.loadCollection(s,[0]);

~validation = {arg batchsize = 30; // -> must be < ~tsize
	var pred = List.new;
	var gtruth = List.new;
	var batchIds = ~testIds.scramble[..batchsize-1]; //every test-batch rand data pts
	var tstLoss;
	{
		batchIds.do({arg id;
			//get input-test-points / then predict, load pred to fa and get gTruth pts and load to fa
			~inTest.getPoint(id.asString,~predSourceBuf,action:
				{
					~mlp.predictPoint(~predSourceBuf,~predTargetBuf,action:{
						~predTargetBuf.loadToFloatArray(action:{
							arg pr; pred.add(pr).flat;
							"*".post;
							~outTest.getPoint(id.asString,~groundTruthBuf,action:{
								~groundTruthBuf.loadToFloatArray(action:{
									arg gtr; gtruth.add(gtr).flat;
									"_".post;
								});
							});
						});
					});
				};
			);
			0.1.wait;
		});
		0.2.wait;
		"".postln;
		//calc MSE and add to list
		tstLoss = (pred - gtruth).flat.squared.sum / batchsize; // MSError
		~testLoss.add(tstLoss); //add it to plot while training (plot func below)

		"predictions    : % ".format(pred.flat.round(0.0001)).postln;
		"ground truths  : % ".format(gtruth.flat.round(0.0001)).postln;
		"test loss      : % \n ".format(tstLoss.round(0.0001)).postln;
	}.fork;
};

~train = {
	var testbatchSize = 30;
	~mlp.fit(~inTrain,~outTrain,{
		arg error;
		~trainLoss.add(error); //add it to plot while training
		"training loss  : % ".format(error.round(0.0001)).postln;
		fork{
			~validation.(testbatchSize);
			if(~continuously_train){
				(testbatchSize+5* 0.1).wait; // eg 30 points validation for each batch -> 20 * 0.1 .. 2 +1/2 secs, then start again
				~train.()
			};
		};
	});
};
~train.();
)




//~mlp.learnRate_(0.01);
~continuously_train = false; //stop it



//exponential moving average (EMA) func
/*EMA implementation according to https://github.com/tensorflow/tensorboard/blob/34877f15153e1a2087316b9952c931807a122aa7/tensorboard/components/vz_line_chart2/line-chart.ts#L699*/

(
~ema = {arg list, weight=0;
	var last = list[0];
	var smoothed = List.new;
	list.doAdjacentPairs({arg a,b,i;
		var last = a;
		var next = b;
		var debias_weight = 1;
		var smoothed_val;
		last = last * weight + (1 - weight) * next;
		if(weight != 1, {debias_weight = 1 - pow(weight, i)});
		smoothed_val = last / debias_weight;
		smoothed.add(smoothed_val);
	});
	smoothed[1..]; //kills inf at 0
});

// cheap gui to show loss functions while training
(
var win, view, plot, w=300, h=750;
win = Window.new("loss funcs", Rect(800,0,w,h));
plot = Plotter(bounds: Rect(0,0,w,h), parent: win);
{while{~continuously_train}{
	//well validation loss is too noisy to make EMA usefull (MA below)
	plot.value = [~trainLoss, ~ema.(~trainLoss, 0.1), ~testLoss];
	//plot.superpose = true; //superpose if needed
	plot.refresh; 0.5.wait; }
}.fork(AppClock);
win.front;
);


// general problems
/*
sometimes getting test input and test output datapoints gets messy (s async..) like
*_*_*_*_*_*_*_*_*_**__*_*_*_*_*_*_*_*_*_*_*_*_*_*_**__*_*_*_
                  ^  ^                            ^  ^
or:

**_*_*__*_*_***___*_*_*_**_*__*_*_**_*__*_*_**_*__*_*_*_**__
^     ^     ^  ^        ^    ^ etc.


and sometimes the prdictions for test input points return nil and
the MSE calcualtion fails ( [nil, ...] - [0.2, ... ]) because of nil - a number
(then no validation loss is done - also in the plot... :(
i can at the moment live with this but ...

if it gets too often nil (and throws an error, increasing wait dur 0.1 in validation and train func helps)


so its very slow due to using a routine and waiting for the server
(but well, fine waiting for 1 h to trash things,
compared to waiting for days (using other models...)

after comparing some hyprparms one could remove validation.() from training..

*/








(
// moving average func (!! only works when testLoss.size > winsize
~ma = {arg list, winsize = 8;
	(list.size - winsize).collect({arg i;
		var ids = Array.series(winsize,i);
		var avg = list.at(ids).sum / winsize;
		avg;
	});
}
)

// gives a bit more direction to testLoss-noise (whatever this means here..)
~ma.(~testLoss, winsize: 50).plot;
